/**
 * Client.java
 * Ce fichier contient l'implmentation d'un client du protocole.
 * Il dcode les messages reus et utilise un bon nombre de fonctions
 * de rappel afin de communiquer l'information  l'usager.
 */
package client;

import commun.*;

import client.ecouteur.IEcouteur;

import java.io.*;
import java.net.Socket;
import java.util.regex.*;
import java.util.*;

import java.io.IOException;

/**
 * @author Daniel Castonguay
 * Cette classe implmente une sorte de pont entre les commandes que l'on peut envoyer
 * et les rponses que le serveur peut renvoyer. Elle ne contient aucun tat de la
 * connexion plus avanc que la gestion du socket; c'est donc aux utilisateurs de cette
 * classe de stocker et grer tous les canaux et usagers qu'ils connaissent.
 */
public class Client 
{
	// Classes internes
	//
	private class TacheNOOP extends TimerTask
	{
		public void run()
		{
			try
			{
				synchronized (fluxSortant)
				{
					// Pour ne pas passer par le code qui rinitialise le thread
					fluxSortant.write(Constantes.NOOP + Constantes.LINE_SEPARATOR);
					fluxSortant.flush();
				}
			}
			catch (IOException e)
			{
				cancel();
				exceptionRecue(e);
			}
			catch (NullPointerException e) // Ce qui ne devrait pas arriver
			{
				cancel();
				exceptionRecue(e);
			}
		}
	}
	
	//  Variables membres
	//
	private IEcouteur ecouteur = null;
	private Socket socket = null;
	private BufferedWriter fluxSortant = null;
	private boolean socketConnecte = false;
	private Thread filMessagesEntrants = null;
	private Timer temporisateurNOOP = new Timer("HorlogeClient", true); // Dmon!
	private TimerTask tacheNOOP = null;
	private boolean enDeconnexion = false;
	private boolean noopPeriodique;
	
	// Constructeurs
	//
	public Client()
	{
		this(true);
	}
	
	public Client(boolean gardeConnexionEnVie)
	{
		noopPeriodique = gardeConnexionEnVie;
	}
	
	// Mthodes
	//
	public void setEcouteur(IEcouteur nouvelEcouteur)
	{
		// Je sais que ceci n'est pas thread-safe, mais a n'est jamais
		// utilis une fois le client dmarr.
		ecouteur = nouvelEcouteur;
	}
	
	public void connecter(String adresse, int port) throws IOException
	{
		if (!socketConnecte)
		{
			socket = new Socket(adresse, port);
			fluxSortant = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream()));
			
			socketConnecte = true;
			
			filMessagesEntrants = new Thread(new ThreadReceptionMessagesClient(this, socket));
			filMessagesEntrants.start();
		
			if (noopPeriodique)
			{
				tacheNOOP = new TacheNOOP();
				temporisateurNOOP.schedule(tacheNOOP, Constantes.LIMITE_TEMPORISATEUR, Constantes.LIMITE_TEMPORISATEUR);
			}
		}
	}
	
	public void connecter(String adresse) throws IOException
	{
		connecter(adresse, Constantes.PORT_PAR_DEFAUT);
	}
	
	public void deconnecter()
	{
		enDeconnexion = true;
		
		if (tacheNOOP != null)
			tacheNOOP.cancel();
		tacheNOOP = null;
		
		try
		{
			if (fluxSortant != null)
				fluxSortant.close();
		}
		catch (IOException e) {}
		try
		{
			if (socket != null)
				socket.close();
		}
		catch (IOException e) {}
		
		fluxSortant = null;
		socket = null;
		
		try
		{
			if (filMessagesEntrants != null && Thread.currentThread() != filMessagesEntrants)
				filMessagesEntrants.join();
		}
		catch (InterruptedException e) {}
		
		filMessagesEntrants = null;
		socketConnecte = false;
		enDeconnexion = false;
	}
	
	public void authentifier(String nomUtilisateur, String motDePasse) throws IOException, InvalidCommandFormatException
	{
		Methodes.validerNomUtilisateur(nomUtilisateur);
		if (motDePasse != null && !motDePasse.isEmpty())
		{
			Methodes.validerMotDePasse(motDePasse);
			envoyerCommandeDirectement(String.format("%s %s %s", Constantes.LOGIN, nomUtilisateur, motDePasse));
		}
		else
			envoyerCommandeDirectement(String.format("%s %s", Constantes.LOGIN, nomUtilisateur));
	}
	
	public void quitter(String message) throws IOException, InvalidCommandFormatException
	{
		Methodes.validerMessagePossiblementVide(message);
		envoyerCommandeDirectement(Constantes.QUIT + (message.isEmpty() ? "" : " " + message));
	}
	
	public void envoyerMessageVersCanal(String canal, String message) throws IOException, InvalidCommandFormatException
	{
		Methodes.validerCanal(canal);
		Methodes.validerMessage(message);
		envoyerCommandeDirectement(String.format("%s %s %s", Constantes.MSG, canal, message));
	}
	
	public void envoyerMessageVersUsager(String usager, String message) throws IOException, InvalidCommandFormatException
	{
		Methodes.validerAlias(usager);
		Methodes.validerMessage(message);
		envoyerCommandeDirectement(String.format("%s %s %s", Constantes.MSG, usager, message));
	}
	
	public void changerAlias(String nouvelAlias) throws IOException, InvalidCommandFormatException
	{
		Methodes.validerAlias(nouvelAlias);
		envoyerCommandeDirectement(Constantes.ALIAS + " " + nouvelAlias);
	}
	
	public void joindreCanal(String canal) throws IOException, InvalidCommandFormatException
	{
		Methodes.validerCanal(canal);
		envoyerCommandeDirectement(Constantes.JOIN + " " + canal);
	}
	
	public void quitterCanal(String canal, String message) throws IOException, InvalidCommandFormatException
	{
		Methodes.validerCanal(canal);
		Methodes.validerMessagePossiblementVide(message);
		envoyerCommandeDirectement(String.format("%s %s%s", Constantes.PART, canal, (message.isEmpty() ? "" : " " + message)));
	}
	
	public void obtenirInformationUsager(String alias) throws IOException, InvalidCommandFormatException
	{
		Methodes.validerAlias(alias);
		envoyerCommandeDirectement(String.format("%s %s", Constantes.WHOIS, alias));
	}
	
	public void obtenirInformationCanal(String canal) throws IOException, InvalidCommandFormatException
	{
		Methodes.validerCanal(canal);
		envoyerCommandeDirectement(String.format("%s %s", Constantes.WHOISON, canal));
	}
	
	public void envoyerNOOP() throws IOException
	{
		//validerAuthentification();
		envoyerCommandeDirectement(Constantes.NOOP);
	}
	
	public void envoyerCommandeLiterale(String commande) throws IOException, InvalidCommandFormatException
	{
		analyserCommande(commande);
		envoyerCommandeDirectement(commande);
	}
	
	private void envoyerCommandeDirectement(String commande) throws IOException
	{
		validerConnexion();
		
		if (noopPeriodique)
		{
			tacheNOOP.cancel();
			tacheNOOP = new TacheNOOP();
			temporisateurNOOP.schedule(tacheNOOP, Constantes.LIMITE_TEMPORISATEUR, Constantes.LIMITE_TEMPORISATEUR);
		}
		
		synchronized (fluxSortant)
		{
			fluxSortant.write(commande + Constantes.LINE_SEPARATOR);
			fluxSortant.flush();
		}
	}
	
	private void validerConnexion() throws NotConnectedException
	{
		if (!estConnecte())
			throw new NotConnectedException("Non connecte.");
	}
		
	public boolean estConnecte()
	{
		return (socket != null && socketConnecte && fluxSortant != null);
	}
	
	private void analyserCommande(String commande) throws InvalidCommandFormatException 
	{
		if (commande == null)
			throw new InvalidCommandFormatException("Commande nulle.");
		
		StringTokenizer tokens = new StringTokenizer(commande, " \t");
		
		if (tokens.hasMoreTokens())
		{
			String typeCommande = tokens.nextToken();
			
			if (typeCommande.equals(Constantes.LOGIN))
				analyserLogin(commande);
			else if (typeCommande.equals(Constantes.QUIT))
				analyserQuit(commande);
			else if (typeCommande.equals(Constantes.JOIN))
				analyserJoin(commande);
			else if (typeCommande.equals(Constantes.PART))
				analyserPart(commande);
			else if (typeCommande.equals(Constantes.MSG))
				analyserMessage(commande);
			else if (typeCommande.equals(Constantes.ALIAS))
				analyserAlias(commande);
			else if (typeCommande.equals(Constantes.WHOIS))
				analyserWhois(commande);
			else if (typeCommande.equals(Constantes.WHOISON))
				analyserWhoisOn(commande);
			else if (typeCommande.equals(Constantes.NOOP))
				analyserNoop(commande);
			else
				throw new InvalidCommandFormatException("Commande inconnue.");
		}
		else
			throw new InvalidCommandFormatException("Commande nulle.");
	}
	
	private void analyserLogin(String commande) throws InvalidCommandFormatException
	{
		// LOGIN dj valid
		StringTokenizer tokens = new StringTokenizer(commande, " \t");
		
		if (tokens.countTokens() >= 2 && tokens.countTokens() <= 3)
		{
			tokens.nextToken();
			Methodes.validerNomUtilisateur(tokens.nextToken());
			
			if (tokens.countTokens() == 3)
				Methodes.validerMotDePasse(tokens.nextToken());
		}
		else
			throw new InvalidCommandFormatException(Constantes.LOGIN + ": nombre de parametres incorrect.");
	}
	
	private void analyserQuit(String commande) throws InvalidCommandFormatException
	{
		// QUIT dj valid
		Methodes.validerMessagePossiblementVide(commande.substring(Constantes.QUIT.length()).trim());		
	}
	
	private void analyserJoin(String commande) throws InvalidCommandFormatException
	{
		// JOIN dj valid
		StringTokenizer tokens = new StringTokenizer(commande, " \t");
		
		if (tokens.countTokens() == 2)
		{
			tokens.nextToken();
			Methodes.validerCanal(tokens.nextToken());
		}
		else
			throw new InvalidCommandFormatException(Constantes.JOIN + ": nombre de parametres incorrect.");
	}
	
	private void analyserPart(String commande) throws InvalidCommandFormatException
	{
		// PART dj valid
		StringTokenizer tokens = new StringTokenizer(commande, " \t");
		
		if (tokens.countTokens() >= 2)
		{
			tokens.nextToken(); // PART
			String canal = tokens.nextToken(); 
			Methodes.validerCanal(canal);
			Methodes.validerMessagePossiblementVide(commande.substring(commande.indexOf(canal) + canal.length()).trim());
		}
		else
			throw new InvalidCommandFormatException(Constantes.PART + ": nombre de parametres incorrect.");
	}
	
	private void analyserMessage(String commande) throws InvalidCommandFormatException
	{
		// MSG dj valid
		StringTokenizer tokens = new StringTokenizer(commande, " \t");
		
		if (tokens.countTokens() >= 3)
		{
			tokens.nextToken(); // MSG
			String destinataire = tokens.nextToken();
			if (Methodes.estCanal(destinataire))
			{
				Methodes.validerMessage(commande.substring(commande.indexOf(destinataire) + destinataire.length()).trim());
			}
			else if (Methodes.estAlias(destinataire))
			{
				Methodes.validerMessage(commande.substring(commande.indexOf(destinataire, Constantes.MSG.length()) + destinataire.length()).trim());
			}
			else
				throw new InvalidCommandFormatException(Constantes.MSG + ": Format de commande invalide.");
		}
		else
			throw new InvalidCommandFormatException(Constantes.MSG + ": nombre de parametres incorrect.");
	}
	
	private void analyserAlias(String commande) throws InvalidCommandFormatException
	{
		// ALIAS dj valid
		StringTokenizer tokens = new StringTokenizer(commande, " \t");
		
		if (tokens.countTokens() == 2)
		{
			tokens.nextToken();
			Methodes.validerAlias(tokens.nextToken());
		}
		else
			throw new InvalidCommandFormatException(Constantes.ALIAS + ": nombre de parametres incorrect.");
	}
	
	private void analyserWhois(String commande) throws InvalidCommandFormatException
	{
		// WHOIS dj valid
		StringTokenizer tokens = new StringTokenizer(commande, " \t");
		
		if (tokens.countTokens() == 2)
		{
			tokens.nextToken();
			Methodes.validerAlias(tokens.nextToken());
		}
		else
			throw new InvalidCommandFormatException(Constantes.WHOIS + ": nombre de parametres incorrect.");
	}
	
	private void analyserWhoisOn(String commande) throws InvalidCommandFormatException
	{
		// WHOISON dj valid
		StringTokenizer tokens = new StringTokenizer(commande, " \t");
		
		if (tokens.countTokens() == 2)
		{
			tokens.nextToken();
			Methodes.validerCanal(tokens.nextToken());
		}
		else
			throw new InvalidCommandFormatException(Constantes.WHOISON + ": nombre de parametres incorrect.");
	}
	
	private void analyserNoop(String commande) throws InvalidCommandFormatException
	{
		// NOOP dj valid
		StringTokenizer tokens = new StringTokenizer(commande, " \t");
		
		if (tokens.countTokens() != 1)
			throw new InvalidCommandFormatException(Constantes.NOOP + ": trop de parametres.");
	}
	
	private void traiterMessage(String message)
	{
		if (ecouteur != null)
			ecouteur.afficherReponseBrute(message);
		
		Matcher m = null;
		if ((m = Constantes.PATTERN_LOGGED_IN.matcher(message)).matches())
			traiterLoggedIn(m.group(2));
		else if ((m = Constantes.PATTERN_LOGIN_INVALID.matcher(message)).matches())
			traiterLoginInvalid();
		else if ((m = Constantes.PATTERN_ALREADY_LOGGED_IN.matcher(message)).matches())
			traiterAlreadyLoggedIn();
		else if ((m = Constantes.PATTERN_JOINED.matcher(message)).matches())
			traiterJoined(m.group(2), m.group(3).split(" "));
		else if ((m = Constantes.PATTERN_JOIN_REFUSED.matcher(message)).matches())
			traiterJoinRefused(m.group(2), (m.group(4) != null ? m.group(4) : ""));
		else if ((m = Constantes.PATTERN_PARTED.matcher(message)).matches())
			traiterParted(m.group(2));
		else if ((m = Constantes.PATTERN_PARTED_ERROR.matcher(message)).matches())
			traiterPartedError(m.group(2), (m.group(4) != null ? m.group(4) : ""));
		else if ((m = Constantes.PATTERN_ALIAS_CHANGED.matcher(message)).matches())
			traiterAliasChanged(m.group(2));
		else if ((m = Constantes.PATTERN_ALIAS_INVALID.matcher(message)).matches())
			traiterAliasInvalid(m.group(2));
		else if ((m = Constantes.PATTERN_ALIAS_ALREADY_USED.matcher(message)).matches())
			traiterAliasAlreadyUsed(m.group(2));
		else if ((m = Constantes.PATTERN_MSG_NOT_SENT_CANAL.matcher(message)).matches())
			traiterMsgNotSentCanal(m.group(2), (m.group(4) != null ? m.group(4) : ""));
		else if ((m = Constantes.PATTERN_MSG_NOT_SENT_ALIAS.matcher(message)).matches())
			traiterMsgNotSentAlias(m.group(2), (m.group(4) != null ? m.group(4) : ""));
		else if ((m = Constantes.PATTERN_USER.matcher(message)).matches())
			traiterUser(m.group(2), m.group(3), (!m.group(4).isEmpty() ? m.group(4).split(" ") : null));
		else if ((m = Constantes.PATTERN_ALIAS_NOT_FOUND.matcher(message)).matches())
			traiterAliasNotFound(m.group(2));
		else if ((m = Constantes.PATTERN_USERS_ON.matcher(message)).matches())
			traiterUsersOn(m.group(2), m.group(3).split(" "));
		else if ((m = Constantes.PATTERN_CHANNEL_NOT_FOUND.matcher(message)).matches())
			traiterChannelNotFound(m.group(2));
		else if ((m = Constantes.PATTERN_NOT_LOGGED_IN.matcher(message)).matches())
			traiterNotLoggedIn();
		else if ((m = Constantes.PATTERN_EVENT_MSG_ALIAS.matcher(message)).matches())
			traiterEventMsgAlias(m.group(3), m.group(4));
		else if ((m = Constantes.PATTERN_EVENT_MSG_CANAL.matcher(message)).matches())
			traiterEventMsgCanal(m.group(3), m.group(4), m.group(5));
		else if ((m = Constantes.PATTERN_EVENT_JOIN.matcher(message)).matches())
			traiterEventJoin(m.group(3), m.group(4));
		else if ((m = Constantes.PATTERN_EVENT_PART.matcher(message)).matches())
			traiterEventPart(m.group(3), m.group(4), (m.group(6) != null ? m.group(6) : ""));
		else if ((m = Constantes.PATTERN_EVENT_ALIAS.matcher(message)).matches())
			traiterEventAlias(m.group(3), m.group(4));
		else if ((m = Constantes.PATTERN_EVENT_QUIT.matcher(message)).matches())
			traiterEventQuit(m.group(3), (m.group(5) != null ? m.group(5) : ""));
		else if ((m = Constantes.PATTERN_EVENT_TIMEOUT.matcher(message)).matches())
			traiterEventTimeout();
		else if ((m = Constantes.PATTERN_EVENT_DISCONNECTED.matcher(message)).matches())
			traiterEventDisconnected((m.group(4) != null ? m.group(4) : ""));
		else if ((m = Constantes.PATTERN_EVENT_COMMAND_ERROR.matcher(message)).matches())
			traiterEventCommandError();
		else
			traiterReponseNonReconnue(message);
	}
	
	private void traiterLoggedIn(String alias)
	{
		if (ecouteur != null)
			ecouteur.authentificationReussie(alias);
	}
	
	private void traiterLoginInvalid()
	{
		if (ecouteur != null)
			ecouteur.authentificationEchouee();
	}
	
	private void traiterAlreadyLoggedIn()
	{
		if (ecouteur != null)
			ecouteur.authentificationDejaEffectuee();
	}
	
	private void traiterJoined(String canal, String[] participants)
	{
		if (ecouteur != null)
			ecouteur.canalJoint(canal, participants);	
	}
	
	private void traiterJoinRefused(String canal, String message)
	{
		if (ecouteur != null)
			ecouteur.canalNonJoint(canal, message);
	}
	
	private void traiterParted(String canal)
	{
		if (ecouteur != null)
			ecouteur.canalQuitte(canal);
	}
	
	private void traiterPartedError(String canal, String message)
	{
		if (ecouteur != null)
			ecouteur.canalNonQuitte(canal, message);
	}
	
	private void traiterAliasChanged(String alias)
	{
		if (ecouteur != null)
			ecouteur.aliasChange(alias);
	}
	
	private void traiterAliasInvalid(String alias)
	{
		if (ecouteur != null)
			ecouteur.aliasInvalide(alias);
	}
	
	private void traiterAliasAlreadyUsed(String alias)
	{
		if (ecouteur != null)
			ecouteur.aliasDejaUtilise(alias);
	}
	
	private void traiterMsgNotSentCanal(String canal, String message)
	{
		if (ecouteur != null)
			ecouteur.messageCanalNonEnvoye(canal, message);
	}
	
	private void traiterMsgNotSentAlias(String alias, String message)
	{
		if (ecouteur != null)
			ecouteur.messageUsagerNonEnvoye(alias, message);
	}
	
	private void traiterUser(String usager, String alias, String[] canaux)
	{
		if (ecouteur != null)
			ecouteur.informationUsager(usager, alias, canaux);
	}
	
	private void traiterAliasNotFound(String alias)
	{
		if (ecouteur != null)
			ecouteur.aliasNonTrouve(alias);
	}
	
	private void traiterUsersOn(String canal, String[] participants)
	{
		if (ecouteur != null)
			ecouteur.informationCanal(canal, participants);
	}
	
	private void traiterChannelNotFound(String canal)
	{
		if (ecouteur != null)
			ecouteur.canalNonTrouve(canal);
	}
	
	private void traiterNotLoggedIn()
	{
		if (ecouteur != null)
			ecouteur.nonAuthentifie();
	}
	
	private void traiterEventMsgAlias(String alias, String message)
	{
		if (ecouteur != null)
			ecouteur.evenementMessage(alias, message);
	}
	
	private void traiterEventMsgCanal(String canal, String alias, String message)
	{
		if (ecouteur != null)
			ecouteur.evenementMessage(canal, alias, message);
	}
	
	private void traiterEventJoin(String canal, String alias)
	{
		if (ecouteur != null)
			ecouteur.evenementCanalJoint(canal, alias);
	}
	
	private void traiterEventPart(String canal, String usager, String message)
	{
		if (ecouteur != null)
			ecouteur.evenementCanalQuitte(canal, usager, message);
	}
	
	private void traiterEventAlias(String vieilAlias, String nouvelAlias)
	{
		if (ecouteur != null)
			ecouteur.evenementAliasModifie(vieilAlias, nouvelAlias);
	}
	
	private void traiterEventQuit(String alias, String message)
	{
		if (ecouteur != null)
			ecouteur.evenementQuitte(alias, message);
	}
	
	private void traiterEventTimeout()
	{
		if (ecouteur != null)
			ecouteur.evenementDelaiDepasse();
	}
	
	private void traiterEventDisconnected(String message)
	{
		if (ecouteur != null)
			ecouteur.evenementDeconnecte(message);
	}
	
	private void traiterEventCommandError()
	{
		if (ecouteur != null)
			ecouteur.evenementCommandeErronee();
	}
	
	private void traiterReponseNonReconnue(String reponse)
	{
		if (ecouteur != null)
			ecouteur.reponseNonReconnue(reponse);
	}
	
	// Modificateur d'accs par dfaut pour que ThreadReceptionMessagesClient y ait accs.
	void messageRecu(String message)
	{
		traiterMessage(message);
	}
	
	void exceptionRecue(Exception exception)
	{
		if (noopPeriodique)
			tacheNOOP.cancel();
		
		afficherErreur(exception);
	}
	
	void connexionInterrompue()
	{
		if (noopPeriodique)
			tacheNOOP.cancel();
		
		deconnecter();
		
		if (ecouteur != null)
			ecouteur.connexionInterrompue();
	}
	
	private void afficherErreur(Exception exception)
	{
		if (ecouteur != null)
			ecouteur.afficherErreur(exception);
	}
	
	boolean enDeconnexion()
	{
		return enDeconnexion;
	}
}
