package serveur;

import commun.*;
import serveur.composant.*;
import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.net.Socket;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;

/**
 * Classe CommunicationClient
 * Implmente une communication base sur les sockets avec un client.
 */
public class CommunicationClient implements Runnable
{	
	/**
	 * Limite additionnelle (en plus de Constantes.LIMITE_TEMPORISATEUR) pour le timeout des clients
	 */
	private int LIMITE_ADDITIONNELLE = 5000; // millisecondes
	
	private boolean commandeFin = false;
	private String messageFin;
	
	private boolean estPret = false;
	private Object pret = new Object(); // Pour lock
	
	private boolean fermetureVoulue = false;
	
	/**
	 * Interface de communication vers le serveur
	 */
	private final IServeurCallBack serveur;
	
	/**
	 * Socket utilis pour la communication avec le client
	 */
	private final Socket sockClient;
	
	/**
	 * Flux de sortie unique pour ce client
	 */
	private final DataOutputStream outToClient;
	
	/**
	 * Usager connect, aprs que l'authentification ait eu lieu
	 */
	private Usager usager;
	
	/**
	 * Constructeur CommunicationClient, coute des requtes du client.
	 * @param serveur Interface de communication vers le serveur
	 * @param sockClient Socket client
	 */
	public CommunicationClient(IServeurCallBack serveur, Socket sockClient)
		throws IOException
	{
		this.serveur = serveur;
		this.sockClient = sockClient;
		this.outToClient = new DataOutputStream(sockClient.getOutputStream());
	}	
	
	/**
	 * Dconnecte le client prsentement connect en fermant le socket.
	 * @param msg Message explicatif
	 */
	public void deconnecterClient(String msg)
	{
		try {
			ecrireMessage(Methodes.construireMessage( (msg != null) ?
				new String[] { Constantes.EVENT, Constantes.DISCONNECTED, msg } :
				new String[] { Constantes.EVENT, Constantes.DISCONNECTED }	));
		}
		catch (IOException e) { 
			// Rien  faire
		}
		fermetureVoulue = true;
		fermerSocket();
	}
	
	/**
	 * Attente jusqu' ce que les donnes puissent tre lues du socket (par un thread).
	 * Cette mthode est toujours appele de l'extrieur.
	 */
	public boolean attendrePret(long timeout)
	{
		synchronized(pret) {
			if (! estPret) {
				try {
					pret.wait(timeout);
				}
				catch (InterruptedException e) {
					// Rien  faire
				}
			}
			return estPret;
		}
	}
	
	/**
	 * Lit, traite et rpond aux commandes envoyes par le client.
	 */
	public void run()
	{	
		if (sockClient == null || sockClient.isClosed())
			return;
			
		try
		{						
			sockClient.setSoTimeout(Constantes.LIMITE_TEMPORISATEUR + LIMITE_ADDITIONNELLE);
			
			synchronized(pret) {
				estPret = true;
				pret.notify();
			}			
			
			BufferedReader inFromClient = new BufferedReader(
								new InputStreamReader(sockClient.getInputStream()));

			String commandeClient = null;
			String reponse = null;
			do
			{
				commandeClient = inFromClient.readLine();
				if (commandeClient == null)
					commandeFin = true;
				else
				{
					// Traitement squentiel pour l'instant
					reponse = traiterCommande(commandeClient);
					
					if (reponse != null)
						outToClient.writeBytes(reponse);
				}
			}
			while (! commandeFin);
		}
		catch (SocketException e)
		{
			if (! fermetureVoulue)
				serveur.notificationException(this, e);
		}
		catch (SocketTimeoutException e)
		{
			serveur.notification(this, "Timeout de l'usager " + assemblerInfosUsager());
			try {
				ecrireMessage(Methodes.construireMessage(
								new String[] { Constantes.EVENT, Constantes.TIMEOUT }));
			}
			catch (IOException ex) {
				// Rien  faire
			}
		}
		catch (Exception e)
		{
			serveur.notificationException(this, e);
		}
		finally
		{
			// Dconnexion du client
			serveur.deconnexion(this, usager, messageFin);
			
			// Fermeture du socket
			fermerSocket();
		}
	}
	
	/**
	 * criture synchronise d'un message dans le flux de sortie.
	 * @param msg Message  transmettre
	 * @throws IOException
	 */
	private void ecrireMessage(String msg)
		throws IOException
	{
		synchronized(outToClient)
		{
			outToClient.writeBytes(msg);
		}
	}
	
	/**
	 * Traite une certaine commande.
	 * @param commandeClient Commande envoye par le client
	 * @return Rponse  la commande
	 */
	private String traiterCommande(String commandeClient)
	{
		StringTokenizer s = new StringTokenizer(commandeClient);
		String reponse = null;

		try {
			// Code de la commande
			String cmd = s.nextToken();
			
			if (cmd.equals(Constantes.QUIT)) /* QUIT */
			{
				int index = commandeClient.indexOf(Constantes.QUIT) + Constantes.QUIT.length() + 1;
				String msg = (index >= commandeClient.length()) ? null : commandeClient.substring(index);
				reponse = traiterQuit(msg);
			}
			else if (cmd.equals("LOGIN")) /* LOGIN */
			{
				reponse = traiterLogin(s.nextToken(), 
						  			s.hasMoreTokens() ? s.nextToken() : null);
			}
			else if (cmd.equals(Constantes.NOOP)) /* NOOP */
			{
				reponse = traiterNoop();
			}
			else
			{
				if (usager == null)
				{
					reponse = traiterLoginRequis();
				}
				else
				{
					if (cmd.equals(Constantes.MSG)) /* MSG */
					{
						String destination = s.nextToken();
						reponse = traiterMsg(destination,
								   commandeClient.substring(commandeClient.indexOf(destination) + 
										destination.length() + 1));
					}
					else if (cmd.equals(Constantes.JOIN)) /* JOIN */
					{
						reponse = traiterJoin(s.nextToken());				
					}
					else if (cmd.equals(Constantes.PART)) /* PART */
					{
						String canal = s.nextToken();
						int index = commandeClient.indexOf(canal) + canal.length() + 1;
						reponse = traiterPart(canal,
											  (index >= commandeClient.length()) ? null : commandeClient.substring(index));
					}
					else if (cmd.equals(Constantes.WHOISON)) /* WHOISON */
					{
						reponse = traiterWhoIsOn(s.nextToken());
					}
					else if (cmd.equals(Constantes.WHOIS)) /* WHOIS */
					{
						reponse = traiterWhoIs(s.nextToken());
					}
					else if (cmd.equals(Constantes.ALIAS)) /* ALIAS */
					{
						reponse = traiterAlias(s.nextToken());
					}
					else /* Commande invalide */
					{
						reponse = traiterCommandeInvalide(cmd);
					}
				}			
			}
		}
		catch (NoSuchElementException e)
		{
			reponse = traiterCommandeInvalide("");
		}
		return reponse;
	}
	
	/**
	 * Traite la commande QUIT.
	 * @param msgQuit Message (facultatif) envoy par l'utilisateur lors de la commande.
	 * @return null
	 */
	private String traiterQuit(String msgQuit)
	{
		commandeFin = true;
		messageFin = msgQuit;
		return null;
	}
	
	/**
	 * Traite la commande LOGIN.
	 * @param nomUtilisateur Nom de l'utilisateur
	 * @param motDePasse Mot de passe de l'utilisateur
	 * @return Rponse  la commande
	 */
	private String traiterLogin(String nomUtilisateur, String motDePasse)
	{
		if (usager != null)
			return Methodes.construireMessage(Constantes.ALREADY_LOGGED_IN);
		
		try {
			IDeconnexion callback = new IDeconnexion()
					{
					public void deconnecter(String raison) { deconnecterClient(raison); } 
					};
			
			usager = serveur.authentification(nomUtilisateur,
											  motDePasse,
											  outToClient,
											  callback);
			if (usager == null)
				return Methodes.construireMessage(Constantes.LOGIN_INVALID);
			
			serveur.notification(this, "Authentification de l'usager " + assemblerInfosUsager());
			
			return Methodes.construireMessage(new String[] {Constantes.LOGGED_IN, usager.lireAlias()});
		}
		catch (ServeurException e) {
			return Methodes.construireMessage(Constantes.LOGIN_INVALID);
		}
	}
	
	/**
	 * Traite le cas o une commande ncessitant un login est reue alors
	 * que le client n'est pas connect.
	 * @return Rponse  la commande
	 */
	private String traiterLoginRequis()
	{
		return Methodes.construireMessage(Constantes.NOT_LOGGED_IN);
	}
	
	/**
	 * Traite une demande d'envoi de message MSG.
	 * @param destination Destination du message
	 * @param message Message  envoyer
	 * @return Rponse  la commande
	 */
	private String traiterMsg(String destination, String message)
	{
		try {
			serveur.envoyerMessage(usager, destination, message);
			return null;
		}
		catch (ServeurException e) {
			return Methodes.construireMessage(
					new String[] {Constantes.MSG_NOT_SENT, destination, e.getMessage()});
		}
	}
	
	/**
	 * Traite la commande JOIN.
	 * @param canal Nom du canal  joindre.
	 * @return Rponse  la commande
	 */
	private String traiterJoin(String canal)
	{
		try {
			Usager[] usagers = serveur.joindreCanal(usager, canal);
			
			String listeUsagers = construireListeUsagers(usagers);
			
			return Methodes.construireMessage(
					new String[] {Constantes.JOINED, canal, listeUsagers});
			
		}
		catch (ServeurException e) {
			return Methodes.construireMessage(
					new String[] {Constantes.JOIN_REFUSED, canal, e.getMessage()});
		}
	}
	
	/**
	 * Traite la commande PART.
	 * @param canal Nom du canal  quitter
	 * @param msgPart Message lors du dpart
	 * @return Rponse  la commande
	 */
	private String traiterPart(String canal, String msgPart)
	{
		try {
			serveur.quitterCanal(usager, canal, msgPart);
			return Methodes.construireMessage(
					new String[] {Constantes.PARTED, canal});
		}
		catch (ServeurException e) {
			return Methodes.construireMessage(
					new String[] {Constantes.PARTED_ERROR, canal, e.getMessage()});
		}
	}
	
	/**
	 * Traite la commande WHOISON.
	 * @param canal Nom du canal
	 * @return Liste des utilisateurs dans le canal
	 */
	private String traiterWhoIsOn(String canal)
	{
		try {
			Usager[] usagers = serveur.getListeUsagers(canal);
			String listeUsagers = construireListeUsagers(usagers);
			
			return Methodes.construireMessage(
					new String[] {Constantes.USERS_ON, canal, listeUsagers});
		}
		catch (ServeurException e) {
			return Methodes.construireMessage(
					new String[] {Constantes.CHANNEL_NOT_FOUND, canal});
		}
	}
	
	/**
	 * Traite la commande WHOIS.
	 * @param alias Alias courant de l'usager
	 * @return
	 */
	private String traiterWhoIs(String alias)
	{
		try {
			Usager usager = serveur.getInfoUsager(alias);
			
			String listeCanaux = construireListeCanaux(usager.listeCanaux());			
			
			return Methodes.construireMessage(
					new String[] {Constantes.USER, usager.lireNomUsager(), usager.lireAlias(), listeCanaux});
		}
		catch (ServeurException e) {
			return Methodes.construireMessage(
					new String[] {Constantes.ALIAS_NOT_FOUND, alias});
		}
	}
	
	/**
	 * Traite la commande ALIAS.
	 * @param nouvelAlias Nouvel alias dsir de l'utilisateur
	 * @return Rponse  la commande
	 */
	private String traiterAlias(String nouvelAlias)
	{
		try {
			serveur.changerAlias(usager, nouvelAlias);
			return Methodes.construireMessage(
					new String[] {Constantes.ALIAS_CHANGED, nouvelAlias});
		}
		catch (ServeurException e) {
			return Methodes.construireMessage(
					new String[] {Constantes.ALIAS_ALREADY_USED, nouvelAlias});
		}
		catch (InvalidCommandFormatException e) {
			return Methodes.construireMessage(
					new String[] {Constantes.ALIAS_INVALID, nouvelAlias});
		}
	}
	
	/**
	 * Traite la commande NOOP.
	 * @return null
	 */
	private String traiterNoop()
	{
		// Rien  faire (NOOP est utilise pour rinitialiser le timer)
		return null;
	}
	
	/**
	 * Traite une commande invalide.
	 * @param cmd Commande invalide
	 * @return Rponse  la commande
	 */
	private String traiterCommandeInvalide(String cmd)
	{
		// 'cmd' n'est pas utilis
		return Methodes.construireMessage(Constantes.COMMAND_ERROR);
	}
	
	/**
	 * Construit une chane de caractres reprsentant un ensemble d'usagers.
	 * @param usagers Liste d'usagers
	 * @return Chane de caractres reprsentant les usagers
	 */
	private String construireListeUsagers(Usager[] usagers)
	{
		StringBuilder listeUsagers = new StringBuilder();
		for (Usager usager : usagers)
		{
			listeUsagers.append(usager.lireAlias()).append(" ");
		}
		return listeUsagers.toString();
	}
	
	/**
	 * Construit une chane de caractres reprsentant un ensemble de canaux.
	 * @param canaux Liste de canaux
	 * @return Chane de caractres reprsentant les canaux
	 */
	private String construireListeCanaux(Canal[] canaux)
	{
		StringBuilder listeCanaux = new StringBuilder();
		for (Canal canal : canaux)
		{
			listeCanaux.append(canal.lireNom()).append(" ");
		}
		return listeCanaux.toString();
	}
	
	/**
	 * Ferme le socket client (si celui-ci est ouvert).
	 */
	private void fermerSocket()
	{
		try {
			if (sockClient != null && !sockClient.isClosed())
			{
				sockClient.close();
				serveur.notification(this, "Dconnexion de l'usager " + assemblerInfosUsager());
			}
		}
		catch (IOException e) {
			serveur.notificationException(this, e);
		}
	}
	
	/**
	 * 
	 * @return Chane de caractres formate reprsentant les infos de l'usager
	 */
	private String assemblerInfosUsager()
	{
		return ((usager != null) ? ("'" + usager + "'") : ("")) + 
				"  l'adresse " + sockClient.getInetAddress() + " port " + sockClient.getPort();
	}
}
