package serveur;

import commun.*;
import java.io.*;
import java.net.Socket;
import java.util.*;
import java.util.concurrent.*;
import java.util.StringTokenizer;
import serveur.composant.*;
import serveur.notification.*;

// TODO : revoir la synchronisation, genre usager join un canal qui vient d'tre dtruit...

/**
 * Classe Serveur, centralise la gestion des canaux et des utilisateurs.
 */
public class Serveur implements IServeurCallBack {
		
	private static final int TAILLE_MAXIMALE_POOL = 0; // Par dfaut aucune limite
	private static final long ATTENTE_THREAD_DISPONIBLE = 1000;  // millisecondes
	
	private static final String MSG_DECONNEXION = "Dconnexion automatique";
	
	/**
	 * couteur  notifier
	 */
	private IServeurEcouteur ecouteur = null;
	
	private Object ecouteurLock = new Object(); // Lock
	
	/**
	 * Log des vnements
	 */
	private PrintWriter log = null;
	
	/**
	 * Fichier utilis pour stocker l'ensemble des comptes
	 */
	private File fichierComptes = null;
	
	/**
	 * Numro de port serveur
	 */
	private int noPort = Constantes.PORT_PAR_DEFAUT;
	
	/**
	 * Nom maximal de connexions concurrentes
	 */
	private int maxConcurrentConnexions = TAILLE_MAXIMALE_POOL;
	
	/**
	 * Flag indiquant si le serveur a dmarr
	 */
	private boolean estDemarre = false;
	
	/**
	 * Pool de threads
	 */
	private ExecutorService threadPool;
	
	/**
	 * Thread charg de l'coute des demandes de connexion
	 */
	private ThreadSocketConnexion threadConn;
	
	/**
	 * Liste des sockets clients encore ouverts
	 */
	private final Set<CommunicationClient> socketsOuverts;
	
	/**
	 * Mapping des noms de canaux vers les instances
	 */
	private final Map<String, Canal> canaux;
	
	/**
	 * Mapping des noms uniques vers les instances
	 */
	 
	private final Map<String, Usager> usagers;
	
	/**
	 * Mapping des alias vers les instances
	 */
	private final Map<String, Usager> alias;
	
	/**
	 * Mapping des noms d'usager vers les mots de passe
	 */
	private Map<String, String> comptes;
	
	/**
	 * Constructeur, initialise les collections.
	 */
	public Serveur()
	{
		canaux = Collections.synchronizedMap(new HashMap<String, Canal>());
		usagers = Collections.synchronizedMap(new HashMap<String, Usager>());
		alias = Collections.synchronizedMap(new HashMap<String, Usager>());
		
		socketsOuverts = Collections.synchronizedSet(new HashSet<CommunicationClient>());
	}
	
	/**
	 * @param maxConcurrentconnexions Nombre maximal de connexions concurrentes acceptes, 0 si aucune limite
	 */
	public void assignerMaxConnexionsConcurrentes(int maxConcurrentConnexions)
		throws ArgumentFormatException
	{
		if (maxConcurrentConnexions < 0)
			throw new ArgumentFormatException("Nombre maximal de connexions doit tre plus grand ou gal  0.");
		this.maxConcurrentConnexions = maxConcurrentConnexions;
	}
	
	/**
	 * Assigne le port sur lequel couter les requtes.
	 * @param noPort Numro du port
	 */
	public void assignerPort(int noPort)
	{
		this.noPort = noPort;
	}
	
	/**
	 * Assigne l'couteur utilis pour l'envoi de notifications.
	 * @param ecouteur
	 */
	public void assignerEcouteur(IServeurEcouteur nouvelEcouteur)
	{
		synchronized(ecouteurLock) {
			ecouteur = nouvelEcouteur;
		}
	}
	
	/**
	 * Assigne les diffrents comptes possibles (pour l'authentification).
	 * @param comptes Mapping des noms d'usagers vers les mots de passe
	 */
	public void assignerComptes(Map<String, String> comptes)
	{
		this.comptes = comptes;
	}
		
	/**
	 * Assigne les diffrents comptes possibles (pour l'authentification).
	 * @param fichier Fichier contenant sur chaque ligne un nom d'usager et un mot de passe spars par un espace
	 */
	public void assignerComptes(File fichier)
	{
		fichierComptes = fichier;
		comptes = new TreeMap<String, String>();
		
		try {
			// Bienvenue dans l'incomprhensible monde des I/O de java
		    FileInputStream fstream = new FileInputStream(fichier);
		    DataInputStream in = new DataInputStream(fstream);
		    InputStreamReader sr = new InputStreamReader(in);
		    BufferedReader br = new BufferedReader(sr);
		    // 45 allocations plus tard...
		    
		    String strLine;
		    while ((strLine = br.readLine()) != null) {
		    	StringTokenizer st = new StringTokenizer(strLine);
		    	if (st.hasMoreTokens())
		    	{
		    		String nomUsager = st.nextToken();
		    		String motDePasse = (st.hasMoreTokens() ? st.nextToken() : "");
		    		comptes.put(nomUsager, motDePasse);
		    	}
		    }
		    in.close();
		}
		catch (Exception e) {
			notification(this, "Une erreur a eu lieu lors de la lecture du fichier de comptes.");
		}	
	}
	
	/**
	 * Ajoute un compte  la liste des comptes disponibles. Si un fichier de comptes existe, ajoute
	 * ce compte au fichier.
	 * @param nomUsager Nom d'usager
	 * @param motDePasse Mot de passe
	 * @throws ServeurException
	 */
	public void ajouterCompte(String nomUsager, String motDePasse)
		throws ServeurException
	{
		if (comptes.containsKey(nomUsager))
			throw new ServeurException("Le nom d'usager '" + nomUsager + "' existe dj.");
		
		comptes.put(nomUsager, motDePasse);
		
		try {
			if (fichierComptes != null)
			{
				FileWriter fw = new FileWriter(fichierComptes, true);	
				PrintWriter out = new PrintWriter(fw);
				out.println(nomUsager + ((motDePasse != null) ? " " + motDePasse : ""));
				out.close();
			}
		}
		catch (Exception e) {
			// Rien  faire (service non essentiel de toute faon)
		}
	}
	
	/**
	 * Dconnecte l'usager spcifi du serveur.
	 * @param nomUsager Nom de l'usager  expulser
	 * @param msgPersonnel Message  envoyer  l'usager lors de la dconnexion
	 * @throws ServeurException
	 */
	public void expulserUsager(String nomUsager, String msgPersonnel)
		throws ServeurException
	{
		Usager usager = usagers.get(nomUsager);
		if (usager == null)
			throw new ServeurException("L'usager '" + nomUsager + "' n'est pas connect.");
		
		usager.deconnecter(msgPersonnel);
		traiterDeconnexion(usager, null);
		notification(this, "Expulsion de l'usager '" + usager + "' avec le message : '" + msgPersonnel + "'.");
	}
	
	/**
	 * @return Liste des canaux prsentement ouverts
	 */
	public Canal[] listeCanaux()
	{
		return canaux.values().toArray(new Canal[] {});
	}
	
	/**
	 * @return Liste des usagers prsentement connects
	 */
	public Usager[] listeUsagers()
	{
		return usagers.values().toArray(new Usager[] {});
	}
	
	/**
	 * Dmarre le serveur.
	 * @throws ServeurException
	 */
	public void demarrer()
		throws ServeurException
	{
		if (estDemarre)
			throw new ServeurException("Serveur est dj dmarr.");
		
		ouvrirFichierLog();
		creerThreadPool();
		try {
			demarrerEcouteSocket();
		}
		catch (IOException e) {
			throw new ServeurException(e.getMessage());
		}
		
		notification(this, "Serveur dmarr.");
		estDemarre = true;
	}
	
	/**
	 * Arrte le serveur.
	 * @throws ServeurException
	 */
	public void arreter()
		throws ServeurException
	{
		if (! estDemarre)
			throw new ServeurException("Serveur n'a pas encore dmarr.");
		
		arreterEcouteSocket();
		
		deconnecterUsagers();
		arreterThreadPool();
		
		notification(this, "Serveur arrt.");
		estDemarre = false;
		
		fermerFichierLog();
	}
	
	/**
	 * Cre le fichier journal.
	 */
	private void ouvrirFichierLog()
	{
		try {
			String nomFichierLog = String.valueOf(new Date().getTime()) + ".txt";
			
			FileWriter fw = new FileWriter(nomFichierLog);
			this.log = new PrintWriter(fw);
		}
		catch (IOException e) {
			notificationException(this, e);
		}
	}
	
	/**
	 * Ferme le fichier journal.
	 */
	private void fermerFichierLog()
	{
		if (log != null)
			log.close();
	}
	
	/**
	 * Cre et initialise le pool de threads (qui seront utiliss pour la communication avec les clients).
	 */
	private void creerThreadPool()
	{
		//return new ThreadPoolExecutor(CORE_POOL_SIZE, 
		//							  MAXIMUM_POOL_SIZE, 
		//							  KEEP_ALIVE_TIME,
		//							  TimeUnit.MILLISECONDS,
		//							  new SynchronousQueue <Runnable>()/*,
		//							  new ServeurThreadFactory()*/);
		
		if (maxConcurrentConnexions == 0)
			threadPool = Executors.newCachedThreadPool();
		else
			threadPool =  Executors.newFixedThreadPool(maxConcurrentConnexions);
	}
	
	/**
	 * Dmarre le thread d'coute du socket serveur.
	 * @throws IOException
	 */
	private void demarrerEcouteSocket()
		throws IOException
	{
		threadConn = new ThreadSocketConnexion(this, noPort);
		threadConn.start();
	}
	
	/**
	 * Arrte le pool de threads.
	 */
	private void arreterThreadPool()
	{
		threadPool.shutdownNow();
	}
	
	/**
	 * Tente de dconnecter (en douceur) les clients connects.
	 */
	private void deconnecterUsagers()
	{
		synchronized(usagers) {
			synchronized(alias) {
				synchronized(socketsOuverts) {				
					Iterator<CommunicationClient> clients = socketsOuverts.iterator();
					while (clients.hasNext())
					{
						clients.next().deconnecterClient("Arrt du serveur");
					}
				}
			}
		}
	}
	
	/**
	 * Arrte l'coute du socket serveur.
	 */
	private void arreterEcouteSocket()
	{
		if (threadConn != null) 
		{
			threadConn.signalFermeture();
		}
	}
	
	/**
	 * Recherche d'une entit de communication.
	 * @param destination Nom de l'entit
	 * @return Entit
	 * @throws ServeurException
	 */
	private ICommunication trouverEntite(String destination)
		throws ServeurException
	{
		ICommunication entite = null;
		
		if (destination.charAt(0) == '#')
		{
			entite = canaux.get(destination);
			if (entite == null)
				throw new ServeurException("Canal n'existe pas.");
		}
		else
		{
			entite = alias.get(destination);
			if (entite == null)
				throw new ServeurException("Usager non connect.");
		}
		return entite;
	}
	
	/**
	 * Traite la dconnexion en retirant l'usager des structures de donnes concernes.
	 * @param usager Usager  supprimer
	 * @param message Message de dconnexion laiss par l'usager
	 */
	private void traiterDeconnexion(Usager usager, String message)
	{
		Canal[] canaux = usager.listeCanaux();
		for (int i=0;i<canaux.length;i++)
			retirerUsagerDuCanal(canaux[i], usager);
		
		synchronized(usagers) {
			if (usagers.get(usager.lireNomUsager()) == usager)
				usagers.remove(usager.lireNomUsager());
		}
		
		alias.remove(usager.lireAlias());
		
		notificationAuxCanaux(usager, new NotificationDeconnexion(usager.lireAlias(), message));
	}
	
	/**
	 * Parcourt tous les canaux auxquels l'usager est connect et notifie (une fois seulement) chaque
	 * usager prsent dans un ou plusieurs canaux commun(s).
	 * @param usager Usager
	 * @param action Action de notification
	 */
	private void notificationAuxCanaux(Usager usager, NotificationUsager action)
	{
		// Utilisation d'un ensemble pour s'assurer qu'un usager n'est pas averti deux fois
		Set<Usager> usagersAvertis = new TreeSet<Usager>();
		
		Canal[] canaux = usager.listeCanaux();
		for (int i=0;i<canaux.length;i++)
		{
			Usager[] usagers = canaux[i].lireUsagersConnectes();
			for (int j=0;j<usagers.length;j++)
			{
				if (usagersAvertis.add(usagers[j]))
					action.notifierUsager(usagers[j]);
			}
		}
	}
	
	/**
	 * Retire un usager d'un canal, dtruit le canal si celui-ci devient vide.
	 * @param canal
	 * @param usager
	 * @return vrai si l'usager tait prsent dans le canal
	 */	
	private boolean retirerUsagerDuCanal(Canal canal, Usager usager)
	{
		boolean succes = canal.retirerUsager(usager);
		if (succes) {
			// Synchronis afin de garantir l'intgrit du canal au cours du test de la condition.
			synchronized (canaux) {
				if (canal.estVide())
					canaux.remove(canal.lireNom());
			}
		}
		return succes;
	}
	
	/**
	 * Vrifie les informations d'authentification passes en paramtre.
	 * @param nomUsager Nom d'usager
	 * @param motDePasse Mot de passe
	 * @return Vrai si l'authentification a russi
	 */
	private boolean verifierLogin(String nomUsager, String motDePasse)
	{
		if (motDePasse == null)
			motDePasse = "";
		
		if (comptes == null)
			return true;
		else
		{
			String mdp = comptes.get(nomUsager);
			return (mdp != null && mdp.equals(motDePasse));
		}
	}
	
	/* Implmentation IServeurCallBack */
	
	public void notification(Object emetteur, String notification)
	{
		if (log != null) {
			synchronized(log) {
				log.println(notification);
				log.flush();
			}
		}
		synchronized (ecouteurLock) {
			if (ecouteur != null)
				ecouteur.notification(emetteur, notification);
		}
	}
	
	public void notificationException(Object emetteur, Exception e)
	{
		if (log != null) {
			synchronized (log) {
				e.printStackTrace(log);
				log.flush();
			}
		} 
		synchronized (ecouteurLock) {
			if (ecouteur != null)
				ecouteur.notificationException(emetteur, e);
		}
	}
	
	public void nouvelleConnexionEntrante(Socket sockClient)
	{		
		try {
			CommunicationClient comm = new CommunicationClient(this, sockClient);
			
			// Garde une trace de la commande afin de pouvoir l'annuler ultrieurement si ncessaire.
			Future<?> future = threadPool.submit(comm);
	
			if (comm.attendrePret(ATTENTE_THREAD_DISPONIBLE))
			{
				socketsOuverts.add(comm);
			}
			else
			{
				comm.deconnecterClient("Trop de connexions");
				future.cancel(false);
			}
		}
		catch (IOException e) {
			notificationException(this, e);
		}
	}
	
	public Usager authentification(String nomUsager, String motDePasse, DataOutputStream out, 
									IDeconnexion fctDeconnexion)
		throws ServeurException
	{				
		
		if (! verifierLogin(nomUsager, motDePasse))
			return null;
						
		// Synchronis pour viter l'authentification concurrente d'un mme usager.
		Usager ancienUsager = null, usager = null;
		synchronized(usagers)
		{
			usager = new Usager(nomUsager, out, fctDeconnexion);
			ancienUsager = usagers.put(usager.lireNomUsager(), usager);
		}
		
		if (ancienUsager != null)
		{
			ancienUsager.deconnecter(MSG_DECONNEXION);
			traiterDeconnexion(ancienUsager, MSG_DECONNEXION);
		}
		
		// Le processus d'obtention d'un alias doit galement tre synchronis.
		synchronized(alias)
		{
			String aliasUsager = nomUsager;
			if (alias.containsKey(nomUsager))
			{
				int noCopie=2;
				while (alias.containsKey(aliasUsager = (nomUsager + noCopie)))
					noCopie++;
			}
			usager.modifierAlias(aliasUsager);
			alias.put(aliasUsager, usager);
		}	
		
		return usager;
	}
	
	public void envoyerMessage(Usager source, String destination, String msg) 
		throws ServeurException
	{
		ICommunication destinataire = trouverEntite(destination);
		
		destinataire.envoieMessage(source.lireAlias(), msg);
	}
	
	public Usager[] joindreCanal(Usager usager, String nomCanal)
		throws ServeurException
	{
		// Synchronis pour viter la cration concurrente de deux canaux ayant le mme nom
		Canal canal = null;
		synchronized(canaux)
		{
			canal = canaux.get(nomCanal);
			if (canal == null) {
				try {
					Methodes.validerCanal(nomCanal);
				}
				catch (InvalidCommandFormatException e) {
					throw new ServeurException("Nom de canal incorrect.");
				}
				canal = new Canal(nomCanal);
				canaux.put(nomCanal, canal);
			}
			
			// Le processus d'entre dans le canal doit galement tre synchronis (avec "canaux"),
			// afin d'viter, par exemple, qu'un usager rejoigne un canal dtruit.
			boolean succesCanal = canal.ajouterUsager(usager);
			boolean succesUsager = usager.joindre(canal);
			
			if (!succesCanal && !succesUsager)
				throw new ServeurException("L'usager '" + usager + "' se trouve dj sur le canal '" + nomCanal + "'.");
		}
			
		notification(this, "Usager '" + usager + "' joint le canal '" + nomCanal + "'.");
		
		return canal.lireUsagersConnectes();
	}
	
	public void quitterCanal(Usager usager, String nomCanal, String message)
		throws ServeurException
	{
		Canal canal = canaux.get(nomCanal);
		
		if (canal == null)
			throw new ServeurException("Canal '" + nomCanal + "' n'existe pas.");
		
		boolean succesCanal = retirerUsagerDuCanal(canal, usager);
		boolean succesUsager = usager.quitter(canal);
		
		if (!succesCanal && !succesUsager)
			throw new ServeurException("L'usager '" + usager + "' ne se trouve pas sur le canal '" + 
					nomCanal + "'.");
			
		notification(this, "Usager '" + usager + "' quitte le canal '" + canal.lireNom() + "'.");
	}
	
	public Usager[] getListeUsagers(String nomCanal)
		throws ServeurException
	{
		Canal canal = canaux.get(nomCanal);
		if (canal == null)
			throw new ServeurException("Canal '" + nomCanal + "' n'existe pas.");
		
		return canal.lireUsagersConnectes();
	}

	public Usager getInfoUsager(String aliasUsager)
		throws ServeurException
	{
		Usager usager = alias.get(aliasUsager);
		if (usager == null)
			throw new ServeurException("Alias '" + aliasUsager + "' n'est pas utilis.");
		
		return usager;
	}
	
	public void changerAlias(Usager usager, String nouvelAlias)
		throws InvalidCommandFormatException, ServeurException
	{
		String ancienAlias = null;
		
		Methodes.validerAlias(nouvelAlias);
		
		synchronized (alias) 
		{
			if (alias.containsKey(nouvelAlias))
				throw new ServeurException("Alias dj utilis.");
			
			ancienAlias = usager.lireAlias();
			
			alias.remove(ancienAlias);
			alias.put(nouvelAlias, usager);
		
			notification(this, "Usager '" + usager + "' change son alias pour '" + nouvelAlias + "'.");
			
			usager.modifierAlias(nouvelAlias);		
		}
		
		// Notification au reste de la plante
		notificationAuxCanaux(usager, new NotificationChangementAlias(ancienAlias, nouvelAlias));
	}
	
	public void deconnexion(CommunicationClient comm, Usager usager, String message)
	{
		// Un thread peut appeler cette mthode en call-back lors de la finalisation alors
		// que l'usager a dj t supprim du systme, il faut vrifier si l'usager est valide.
		if (usager != null)
			traiterDeconnexion(usager, message);

		socketsOuverts.remove(comm);
	}
}
