package serveur.composant;

import java.lang.StringBuilder;
import java.util.Collections;
import java.util.Set;
import java.util.TreeSet;
import commun.Constantes;

/**
 * Reprsente un canal  travers lequel tous ses membres peuvent communiquer.
 * 
 * L'ensemble des usagers connects est synchronis.
 * Cet ensemble est accessible seulement  l'interne, les demandes pour y accder retournent un
 * nouveau tableau [] afin d'viter les modifications pendant l'itration (ConcurrentModificationException).
 */
public class Canal implements ICommunication, Comparable<Canal> {
	
	private String nomCanal;
	
	/**
	 * Ensemble des usagers prsentement connects au canal
	 */
	private Set<Usager> usagersConnectes = Collections.synchronizedSet(new TreeSet<Usager>());
	
	/**
	 * Constructeur
	 * @param nomCanal Nom du canal
	 */
	public Canal(String nomCanal)
	{
		this.nomCanal = nomCanal;
	}
	
	/**
	 * @return Vrai si le canal ne contient aucun usager
	 */
	public boolean estVide()
	{
		return usagersConnectes.isEmpty();
	}
	
	/**
	 * @return Nom unique du canal
	 */
	public String lireNom()
	{
		return nomCanal;
	}
	
	/**
	 * Ajoute un usager au canal et notifie les autres usagers du canal.
	 * @param usager Usager  ajouter
	 */
	public boolean ajouterUsager(Usager usager)
	{	
		boolean succes = usagersConnectes.add(usager);

		if (succes)
		{
			// Itrer sur une copie de la liste est ncessaire pour viter les modifications concurrentes.
			Usager[] usagers = usagersConnectes.toArray(new Usager[] {});
			for (int i=0;i<usagers.length;i++)
				if (usagers[i] != usager)
					usagers[i].notificationEntreeUsager(nomCanal, usager.lireAlias());
		}
		return succes;
	}

	/**
	 * Retire l'usager de la liste des usagers connects au canal (aucune notification).
	 * @param usager Usager quittant le canal
	 */
	public boolean retirerUsager(Usager usager)
	{
		return usagersConnectes.remove(usager);
	}
	
	/**
	 * Retire l'usager de la liste des usagers connects au canal.
	 * Envoie une notification de dpart aux autres usagers du canal.
	 * @param usager Usager quittant le canal
	 * @param message Message (peut tre null) de notification envoys aux participants du canal
	 */
	public boolean retirerUsager(Usager usager, String message)
	{
		boolean succes = retirerUsager(usager);
		
		if (succes)
		{
			// Copie de la liste pour viter les modifications concurrentes
			Usager[] usagers = usagersConnectes.toArray(new Usager[] {});
			for (int i=0;i<usagers.length;i++)
				usagers[i].notificationDepartUsager(nomCanal, usager.lireAlias(), message);
		}
		return succes;
	}

	/**
	 * Envoie un message  l'ensemble des utilisateurs du canal.
	 * @param expediteur Expditeur du message
	 * @param message Message envoy aux participants
	 */
	public void envoieMessage(String expediteur, String message)
	{
		Usager[] usagers = usagersConnectes.toArray(new Usager[] {});
		for (int i=0;i<usagers.length;i++)
			usagers[i].envoieMessageProvenantDuCanal(nomCanal, expediteur, message);
	}
	
	/**
	 * @return Liste des usagers connects
	 */
	public Usager[] lireUsagersConnectes()
	{
		return usagersConnectes.toArray(new Usager[] {});
	}
	
	public int compareTo(Canal canal)
	{
		return (this == canal) ? 0 : 1;
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder(lireNom()).append(Constantes.LINE_SEPARATOR);
		for (Usager usager : lireUsagersConnectes())
		{
			sb.append("  ").append(usager).append(Constantes.LINE_SEPARATOR);
		}
		return sb.toString();
	}
}
