package serveur.composant;

import commun.*;
import java.io.DataOutputStream;
import java.io.IOException;
import java.lang.Comparable;
import java.util.Collections;
import java.util.Set;
import java.util.TreeSet;

/**
 * Classe Usager, chaque instance reprsente un usager CONNECT.
 * 
 * L'criture de donnes dans le flux de sortie ainsi que la liste interne de canaux sont synchroniss.
 */
public class Usager implements ICommunication, Comparable<Usager> {

	/**
	 * Flux de sortie pour la communication
	 */
	private final DataOutputStream out;

	/**
	 * Nom unique d'utilisateur
	 */
	private final String nomUsager;
	
	/**
	 * Fonction de dconnexion
	 */
	private final IDeconnexion fctDeconnexion;
	
	/**
	 * Alias par lequel l'usager apparat aux autres usagers
	 */
	private String alias;
	
	/**
	 * Ensemble des canaux auxquels l'usager est connect
	 */
	private Set<Canal> canaux = Collections.synchronizedSet(new TreeSet<Canal>());

	/**
	 * Constructeur d'un usager
	 * @param nomUsager Nom d'usager
	 * @param out Flux de sortie
	 * @param fctDeconnexion Fonction de dconnexion
	 */
	public Usager(String nomUsager, DataOutputStream out, IDeconnexion fctDeconnexion)
	{
		this.nomUsager = nomUsager;
		this.alias = nomUsager;
		this.out = out;
		this.fctDeconnexion = fctDeconnexion;
	}
	
	/**
	 * @return Nom d'usager
	 */
	public String lireNomUsager() 
	{
		return nomUsager;
	}
	
	/**
	 * @return Alias
	 */
	public String lireAlias()
	{
		return alias;
	}
	
	/**
	 * @param nouvelAlias Nouvel alias
	 */
	public void modifierAlias(String nouvelAlias)
	{
		alias = nouvelAlias;
	}
	
	/**
	 * @return Liste des canaux auxquels l'usager est prsentement connect
	 */
	public Canal[] listeCanaux()
	{
		return canaux.toArray(new Canal[] {});
	}
	
	/**
	 * mission d'une notification comme quoi un autre usager a joint un canal.
	 * @param canal Nom du canal en cause
	 * @param usager Usager ayant joint ce canal
	 */
	public void notificationEntreeUsager(String canal, String usager)
	{
		ecrireMessage(new String[] {Constantes.EVENT, Constantes.JOIN, canal, usager} );
	}
	
	/**
	 * mission d'une notification de dpart du canal d'un usager.
	 * @param canal Nom du canal en cause
	 * @param usager Usager ayant quitt ce canal
	 * @param message Message laiss par l'usager
	 */
	public void notificationDepartUsager(String canal, String usager, String message)
	{		
		ecrireMessage((message != null) ? new String[] {Constantes.EVENT, Constantes.PART, canal, usager, message} :
										  new String[] {Constantes.EVENT, Constantes.PART, canal, usager});
	}
	
	/**
	 * mission d'une notification de dconnexion d'un usager.
	 * @param usager Usager s'tant dconnect
	 * @param message Message laiss par l'usager
	 */
	public void notificationDeconnexionUsager(String usager, String message)
	{
		ecrireMessage((message != null) ? new String[] {Constantes.EVENT, Constantes.QUIT, usager, message} :
			  							  new String[] {Constantes.EVENT, Constantes.QUIT, usager});
	}
	
	/**
	 * mission d'une notification de changement d'alias.
	 * @param ancien_alias Ancien alias de l'usager
	 * @param nouvel_alias Nouvel alias de l'usager
	 */
	public void notificationChangementAlias(String ancien_alias, String nouvel_alias)
	{
		ecrireMessage(new String[] {Constantes.EVENT, Constantes.ALIAS, ancien_alias, nouvel_alias} );
	}
	
	/**
	 * Envoi d'un message en provenance d'un canal auquel l'usager est connect.
	 * @param canal Nom du canal
	 * @param expediteur Expditeur du message
	 * @param msg Message
	 */
	public void envoieMessageProvenantDuCanal(String canal, String expediteur, String msg)
	{
		ecrireMessage(new String[] {Constantes.EVENT, Constantes.MSG, canal, expediteur, msg} );
	}
	
	/**
	 * Envoi d'un message personnel.
	 * @param expediteur Expditeur du message
	 * @param msg Message
	 */
	public void envoieMessage(String expediteur, String msg)
	{	
		ecrireMessage(new String[] {Constantes.EVENT, Constantes.MSG, expediteur, msg} );
	}
	
	/**
	 * Joint un canal.
	 * @param canal Canal  joindre
	 */
	public boolean joindre(Canal canal)
	{
		return canaux.add(canal);
	}
	
	/**
	 * Quitte un canal.
	 * @param canal Canal  quitter
	 */
	public boolean quitter(Canal canal)
	{
		return canaux.remove(canal);
	}
	
	/**
	 * Dconnecte l'utilisateur (normalement pas  la suite d'une requte de celui-ci).
	 * @param raison Raison de la dconnexion
	 */
	public void deconnecter(String raison)
	{
		fctDeconnexion.deconnecter(raison);
	}
	
	/**
	 * Envoie un message  l'utilisateur dans le flux de sortie
	 * @param params Parties du message
	 */
	private void ecrireMessage(String[] params)
	{
		try {
			String msg = Methodes.construireMessage(params);
			
			// criture synchronise dans le flux de sortie
			synchronized(out) {
				out.writeBytes(msg);
			}
		}
		catch (IOException e) {
			// Rien  faire
		}
	}
	
	public int compareTo(Usager usager)
	{
		return (this == usager) ? 0 : 1;
	}
	
	@Override
	public String toString() {
		return nomUsager + (alias.equals(nomUsager) ? "" : " (" + alias + ")");
	}
}
