package ui.serveur;

import commun.ArgumentFormatException;
import commun.InvalidCommandFormatException;
import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;
import commun.Constantes;
import serveur.*;
import serveur.composant.*;

/**
 * Classe MainServeur, interface utilisateur pour envoyer des commandes simples au serveur.
 */
public class MainServeur implements IServeurEcouteur {

	static private final String AIDE = "AIDE";
	static private final String LISTE_CANAUX = "LISTECANAUX";
	static private final String LISTE_USAGERS = "LISTEUSAGERS";
	static private final String AJOUTER_COMPTE = "AJOUTERCOMPTE";
	static private final String EXPULSER = "EXPULSER";
	static private final String ARRETER = "ARRETER";
	
	private Serveur serveur;
	
	private boolean commandeFin = false;
	
	/*
	 * Arguments passs en ligne de commande
	 */
	private String nomFichier = null;
	private Integer nMaxConnections = null;
	private Integer nPort = null;
	
	/**
	 * @param args [fichierComptes] [-m maxConnexions] [-p port]
	 */
	public void run(String[] args) {
		
		if (! lireArguments(args))
		{
			afficherUsage();
			return;
		}

		try
		{
			serveur = new Serveur();
			configurerServeur();
		}
		catch (Exception e) {
			System.out.println(e.getMessage());
			return;
		}
		
		InputStreamReader sr = new InputStreamReader(System.in);
		BufferedReader reader = new BufferedReader(sr);
		
		try {			
			serveur.demarrer();
				
			afficherAide();
			
			commandeFin = false;
			while (! commandeFin)
			{
				String commande = reader.readLine();
				try {
					String reponse = traiterCommande(commande);
					if (reponse != null)
						System.out.println(reponse);
				}
				catch (InvalidCommandFormatException e) {
					System.out.println(e.getMessage());
					afficherAide();
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		finally {
			// S'assure de bien arrter le serveur
			try {
				System.out.println("Envoi de la commande d'arrt...");
				serveur.arreter();
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Usage
	 */
	private void afficherUsage()
	{
		System.out.println("Usage: serveur [fichierComptes] [-m maxConnexions] [-p port]");
	}
	
	/**
	 * Liste commandes
	 */
	private void afficherAide()
	{
		System.out.println("Liste des commandes disponibles :");
		System.out.println("  aide");
		System.out.println("  listeCanaux");
		System.out.println("  listeUsagers");
		System.out.println("  ajouterCompte <usager> [mot de passe]");
		System.out.println("  expulser <usager> [message]");
		System.out.println("  arreter");
	}
	
	/**
	 * Extrait les arguments passs en ligne de commande.
	 * @param args Arguments
	 * @return Vrai si les arguments sont valides
	 */
	private boolean lireArguments(String[] args)
	{
		try {
			for (int i=0;i<args.length;i++)
			{
				if (args[i].charAt(0) == '-' && args[i].length() > 1)
				{
					switch (args[i].charAt(1))
					{
					case 'm':
						nMaxConnections = new Integer(Integer.parseInt(args[++i]));
						break;
					case 'p':
						nPort = new Integer(Integer.parseInt(args[++i]));
						break;
					}
				}
				else
				{
					if (nomFichier != null)
						throw new Exception("Argument invalide");
					else
						nomFichier = args[i];
				}
			} 
		}
		catch (Exception e) {
			return false;
		}
		return true;
	}
	
	/**
	 * Configure le serveur  l'aide des variables statiques de configuration.
	 * @throws ArgumentFormatException
	 */
	private void configurerServeur()
		throws ArgumentFormatException
	{
		serveur.assignerEcouteur(this);		

		// Fichier de comptes
		if (nomFichier != null)
		{
			File fichierComptes = new File(nomFichier);
			if (! fichierComptes.exists())
			{
				throw new ArgumentFormatException("Le fichier '" + nomFichier + "' n'existe pas.");
			}
			serveur.assignerComptes(fichierComptes);
		}
			
		// Nombre maximal de connections
		if (nMaxConnections != null)
		{
			serveur.assignerMaxConnexionsConcurrentes(nMaxConnections);
		}
		
		// Port sur lequel le serveur coute
		if (nPort != null)
		{
			serveur.assignerPort(nPort);
		}
	}
	
	/**
	 * Traite la commande reue.
	 * @param commande
	 * @return rponse
	 * @throws InvalidCommandFormatException
	 */
	private String traiterCommande(String commande)
		throws InvalidCommandFormatException
	{
		StringTokenizer s = new StringTokenizer(commande);
		String reponse = null;
		
		try
		{
			String cmd = s.nextToken().toUpperCase();
			
			if (cmd.equals(AIDE))
			{
				afficherAide();
			}
			else if (cmd.equals(ARRETER))
			{
				commandeFin = true;
			}
			else if (cmd.equals(EXPULSER))
			{
				String nomUsager = s.nextToken();
				int index = commande.indexOf(nomUsager) + nomUsager.length() + 1;
				String message = (index >= commande.length()) ? null : commande.substring(index);
							
				try {
					serveur.expulserUsager(nomUsager, message);
					reponse = "L'usager '" + nomUsager + "' a t expuls avec succs.";
				}
				catch (ServeurException e) {
					reponse = e.getMessage();
				}
			}
			else if (cmd.equals(AJOUTER_COMPTE))
			{
				try {
					String nomUsager = s.nextToken();
					serveur.ajouterCompte(nomUsager, (s.hasMoreTokens() ? s.nextToken() : null));
					reponse = "Le compte '" + nomUsager + "' a t ajout avec succs.";
				}
				catch (ServeurException e) {
					reponse = "Erreur : " + e.getMessage();
				}
			}
			else if (cmd.equals(LISTE_CANAUX))
			{
				reponse = construireListeCanaux();
			}
			else if (cmd.equals(LISTE_USAGERS))
			{
				reponse = construireListeUsagers();
			}
			else
			{
				throw new InvalidCommandFormatException("Commande inconnue.");
			}
		}
		catch (NoSuchElementException e)
		{
			throw new InvalidCommandFormatException("Erreur de syntaxe dans la commande.");
		}
		return reponse;
	}
	
	/**
	 * @return Chane de caractres reprsentant l'ensemble des canaux
	 */
	private String construireListeCanaux()
	{
		StringBuilder sb = new StringBuilder("Liste des canaux :").append(Constantes.LINE_SEPARATOR)
									.append("------------------").append(Constantes.LINE_SEPARATOR);
		for (Canal canal : serveur.listeCanaux())
		{
			sb.append(canal).append(Constantes.LINE_SEPARATOR);
		}
		return sb.toString();
	}
	
	/**
	 * @return Chane de caractres reprsentant l'ensemble des usagers
	 */
	private String construireListeUsagers()
	{
		StringBuilder sb = new StringBuilder("Liste des usagers :").append(Constantes.LINE_SEPARATOR)
									.append("-------------------").append(Constantes.LINE_SEPARATOR);
		for (Usager usager : serveur.listeUsagers())
		{
			sb.append(usager);
			Canal[] canaux = usager.listeCanaux();
			for (int i=0;i<canaux.length;i++)
			{
				if (i == 0)
					sb.append(" (");
				sb.append(canaux[i].lireNom());
				
				if (i != (canaux.length-1))
					sb.append(", ");
				else
					sb.append(")");
			}
			sb.append(Constantes.LINE_SEPARATOR);
		}
		return sb.toString();
	}

	/**
	 * Rception d'une notification du serveur.
	 */
	public void notification(Object emetteur, String message)
	{
		synchronized(this) {
			System.out.println(message);
		}
	}
	
	/**
	 * Rception d'une notification d'exception en provenance du serveur.
	 */
	public void notificationException(Object emetteur, Exception e)
	{
		synchronized(this) {
			e.printStackTrace();
		}
		
	}
	
	/**
	 * Main statique
	 * @param args
	 */
	public static void main(String[] args) {
		new MainServeur().run(args);
	}
	
}
